span#logotext2 {
    color: #764099;
}

/* Taken from NumPy */

@import url('https://fonts.googleapis.com/css2?family=Lato:ital,wght@0,400;0,700;0,900;1,400;1,700;1,900&family=Open+Sans:ital,wght@0,400;0,600;1,400;1,600&display=swap');

.navbar-brand img {
   height: 60px;
}
.navbar-brand {
   height: 75px;
}

body {
  font-family: 'Open Sans', sans-serif;
}

pre, code {
  font-size: 100%;
  line-height: 155%;
}

div.output_area div[class*=highlight] pre {
  white-space: pre-wrap;
}

/* Make output lighter gray and no italics for the love of all that is holy! */
html[data-theme="light"] .highlight .go {
  color: #555555;
  font-style: normal;
}

html[data-theme="dark"] .highlight .go {
  font-style: normal;
}

/* OMG why would you bold numbers */
.highlight .mf, .highlight .mi {
  font-weight: 300;
}

/* Override some aspects of the pydata-sphinx-theme: taken from Pandas */

:root {
  /* Use softer blue from bootstrap's default info color */
  /* --pst-color-info: 23, 162, 184; */
  --pst-color-primary: 118, 63, 152;
  --pst-color-success: 40, 167, 69;
  --pst-color-info: 0, 123, 255;
  --pst-color-warning: 255, 193, 7;
  --pst-color-danger: 220, 53, 69;
  --pst-color-text-base: 51, 51, 51;

  --pst-font-size-base: 15px;

  --pst-color-link: 118, 63, 152;
  --pst-color-headerlink-hover: 118, 63, 152;

  /* heading font sizes */
  --pst-font-size-h1: 28px;
  --pst-font-size-h2: 25px;
  --pst-font-size-h3: 20px;
  --pst-font-size-h4: 18px;
  --pst-font-size-h5: 16px;
  --pst-font-size-h6: 15px;
}
